/*
* ATM168.c Code file to test working with ATMEGA168
* Using avr-gcc 4.3.0
*
*/

#undef F_CPU
#define F_CPU 16000000UL	//Your mileage will vary here. See InitMotors() for changes to TCCR0B prescale!
#define BAUD 38400L 
#define BAUD_RR (F_CPU/16L/BAUD-1) 

#include <avr/interrupt.h> 
#include <inttypes.h> 
#include <stdio.h>
#include <avr/io.h>
#include "atm168.h"

/*
 * Enable printf() functionality in avr-gcc libs
 */
FILE *fileDef;
int def_putc(char ch,FILE *fileDef);
int def_getc(FILE *fileDef);

/*
 * Deal with interrupt modified variables.
 */
volatile uint16_t	pTics;				//phase encoder counts (one rotation=180)
volatile uint8_t	direction;			//what is the measured direction?

/*
 * This section deals with the quadrature encoders as interrupts.
 */
ISR(INT0_vect)
/*
 * INT0 vector counting phase 1 encoder rising transitions.
 */
 {
 	pTics++;
 	direction = PHASE2;		//FWD if low, REV if high
 }

ISR(INT1_vect)
/*
 * Yup, the other phase rising transitions.
 */
{
	pTics++;
}

/*
* This code exercises the LEGO servo motor and its encoders
*/
int main(void)
{
		
	initialize();								//Initialize the board
    fileDef = fdevopen(def_putc, def_getc);     //init stdio
	MotorsStop();								// turn off for now
	pTics = 0;
	
	printf("Waiting 10 seconds to time out\n\r");
	waitms(10000);
	
	/*
	 * Go forward and back the same distance to see how the encoders work.
	 * I am using 5cm diameter wheels giving a 15.71cm rollout (circumference).
	 * Going 1000 tics, when we count a tic every 1 degree means 1000/360 = 2.78
	 * revolutions.  2.78 * 15.71 = 43.6cm distance traveled.  We hit it dead on.
	 */
	pTics = 0;
	MotorsGo(100,100);
	while(pTics < 1000);
	MotorsStop();
	pTics = 0;
	MotorsGo(-100,-100);
	while(pTics < 1000);
	MotorsGo(100,100);							//Just to hit the "brakes"
	waitms(100);
	MotorsStop();
	
	/*
	 * Now let's just twiddle the motors back and forth to see the direction changes.
	 */
	while(1)
	{
		PORTBbits.b5 = ~PORTBbits.b5;		//Blink an LED on this port to show we're running
		waitms(1000);
		printf("dir= %d speed = %d\n\r",direction, pTics);
	}
}

void InitMotors(void)
/*
 * Set up the motors and PWM and such.
 */
{
	TCCR0A = 0xF3;			//Fast PWM, clear on match, count up from bottom
	TCCR0B = 0x03;			//prescale by 64 which is a 1KHz PWM more or less on 16MHz
	
	LEFTMOTOR = 0;			//OCR0A
	RIGHTMOTOR = 0;			//OCR0B
}

void initialize()
{
	DDRC = 0xFF; 							// All outputs for now...
	DDRB = 0xFF; 							// Set up the rest according to pinouts
	DDRD = 0xF2;							//INT0,INT1 and RxD inputs, TxD output

	ACSR = 0x80; 							// turn off the comparitor
	
	//Background timer
	TCCR2A = 0x00;							//No PWM
	TCCR2B = 0x05;							//prescale 128 for 16MHz clock
	
	//Use INT0 and INT1 for optical encoders
	EICRA = 0x0F;							//rising edge on INT0 and INT1
	EIMSK = 0x03;							//Turn on both INT0 and INT1

    //Enable UART (BAUD set above)
    UBRR0H = (unsigned char)(BAUD_RR >> 8);
    UBRR0L = (unsigned char)BAUD_RR;
    UCSR0B = (unsigned char)_BV(TXEN0) | _BV(RXEN0);
    UCSR0C = (unsigned char)_BV(UCSZ01) | _BV(UCSZ00);
    
	PORTD = 0x00;
	PORTB = 0x00;
	PORTC = 0x00;							//Initialize all output values
	
	InitMotors();							//Set up the motors
	sei();									//turn on interrupts
}

void SetMotorSpeed(int speedL, int speedR)
/*
 * SetMotorSPeed() works with +/- 255 as values.
 */
{
	static int lastSpeedL = 0;				//Only change speed when it is different
	static int lastSpeedR = 0;
	
	//When we change direction, we also need to change the sense of the PWM line.
	if (speedL != lastSpeedL)
	{
		lastSpeedL = speedL;
		if (speedL >=0)
		{
			L_DIR = FWD;
			TCCR0A &= ~L_NORMALPWM;		//CLear COM0B0		
			LEFTMOTOR = speedL;
		}
		else
		{
			L_DIR = REV;
			TCCR0A |= L_NORMALPWM;		//Set COM0B0 which reverses PWM sense
			LEFTMOTOR = -speedL;
		}
	}
	if (speedR != lastSpeedR)
	{
		lastSpeedR = speedR;
		if (speedR >= 0)
		{
			R_DIR = FWD;
			TCCR0A &= ~R_NORMALPWM;		//Clear COM0A0
			RIGHTMOTOR = speedR;
		}
		else
		{
			R_DIR = REV;
			TCCR0A |= R_NORMALPWM;		//Set COM0A0 which reverses PWM sense
			RIGHTMOTOR = -speedR;
		}
	}
}

void MotorsStop(void)
{
	SetMotorSpeed(0,0);
	M_ENABLE = OFF;						//Turn motor enable line off
}

void MotorsGo(int leftM, int rightM)
{
	SetMotorSpeed(leftM,rightM);
	M_ENABLE = ON;
}

void waitms(int d)
/*
 * This is tuned to 16MHz clock
 */
{
	int delay;

	TCNT2 = 130; 							//(255-130)*8us = about 1ms
	for (delay=0;delay<d;delay++)
	{
		while (TCNT2);
		TCNT2 = 130;
	}
}

/*
 * printf() functions use these.
 */
int def_putc(char ch,FILE *fileDef)
/*
 * The default putc routine for printf(
 */
 {
     while ((unsigned char)(UCSR0A & _BV(UDRE0)) == 0);

     UDR0 = ch;
     return ch;
 }

int def_getc(FILE *fileDef)
{
	//Crude, but workable
	if ((unsigned char)(UCSR0A & (1<<RXC0)))
		return (int)UDR0;
	else
		return (int)-1;
}
